import math
import numpy as np
from pathlib import Path
import json
import os
from tifffile import TiffWriter

class WaveformFileReader:
    fname_zeros = 8
    tiff_ext = "tiff"
    resolution = 88.0
    fireconstants_multiplicationfactor = 4
    software = "BDRC"

    def __init__(self, file):

        # parse the file
        self.fpath = file
        self.fname = Path(file).stem
        self.parse_wavefile(file)

        # pull out the most useful parameters
        self.num_per_row = self._config.samplesPerRow  # 256
        self.stride = self.num_per_row // 4  # 64
        self.image_bin_min = self._config.minFreq  # 16
        self.image_bin_max = self._config.maxFreq  # 120
        self.num_columns = self.image_bin_max - self.image_bin_min
        self.num_channels = self._header.channels  # 8
        self.img_channels = [
            int(i) for i in self.metadata["namesForImageChannels"].keys()
        ]
        self.num_channels_used = len(self.img_channels)
        self.brightfield_channel = self.metadata["lasers"][0]["lightlossIndex"]
        self.pixel_aspect_ratio = self.metadata["lasers"][0]["pixelAspectRatio"]
        self.resolutions = [self.get_horizontal_resolution(), self.resolution]  # X,Y

        # reorder image channels to put brightfield first
        self.img_channels = [self.brightfield_channel] + [
            i for i in self.img_channels if i != self.brightfield_channel
        ]
        self.img_channel_names = [
            self.metadata["namesForImageChannels"][str(i)] for i in self.img_channels
        ]
        self.img_channel_dict = dict(zip(self.img_channels, self.img_channel_names))

    def parse_wavefile(self, fpath):
        # parse the header, config, metadata, and event table
        _header = self.Header()
        _config = self.Header()
        with open(fpath, "rb") as f:
            # from src/net/Projects/Chorus.Analysis/Storage/WaveformFileReader.cs
            _header.identifier = str(f.read(8))
            _header.version = int.from_bytes(f.read(4), "little")
            _header.channels = int.from_bytes(f.read(4), "little")
            _header.channelIndex = int.from_bytes(f.read(4), "little")
            _header.totalChannels = int.from_bytes(f.read(4), "little")
            _header.bitDepth = int.from_bytes(f.read(4), "little")
            _header.compression = int.from_bytes(f.read(4), "little")
            _header.waveformFormat = int.from_bytes(f.read(4), "little")
            _header.guid = f.read(16)
            _header.chunkId = int.from_bytes(f.read(4), "little")
            _header.totalChunks = int.from_bytes(f.read(4), "little")
            _header.eventCount = int.from_bytes(f.read(4), "little")
            _header.eventTablePosition = int.from_bytes(f.read(8), "little")
            _header.waveformStreamPosition = int.from_bytes(f.read(8), "little")
            _header.metadataPosition = int.from_bytes(f.read(8), "little")
            _header.padding = f.read(40)
            _config.length = int.from_bytes(f.read(4), "little")
            _config.identifier = str(f.read(8))
            _config.FIREenabled = int.from_bytes(f.read(4), "little")
            _config.samplesPerRow = int.from_bytes(f.read(4), "little")
            _config.minFreq = int.from_bytes(f.read(4), "little")
            _config.maxFreq = int.from_bytes(f.read(4), "little")
            _config.freqPitch = int.from_bytes(f.read(4), "little")
            _config.padding = f.read(96)
            f.seek(_header.eventTablePosition)
            eventTable = {}
            for eventIndex in range(_header.eventCount):
                # eventIndex is the sequential index of the waveform (including 0-length waveforms, which are events with no image recorded)
                # eventIndex is used by Image Extractor for filenaming
                # eventIndex should correspond to the relative index of an event in the FCS file
                # eventId corresponds to the value in the "EventNumber0" fcs parameter
                # see WaveformFileReader
                eventId = int.from_bytes(f.read(8), "little")
                eventOffset = int.from_bytes(f.read(8), "little")
                eventLength = int.from_bytes(f.read(8), "little")
                if eventLength != 0:
                    eventTable.update({eventId: (eventOffset, eventLength, eventIndex)})
            f.seek(_header.metadataPosition)
            metadata = f.read()
        self._header = _header
        self._config = _config
        self.eventTable = eventTable
        self.eventIds = list(eventTable.keys())
        self.eventIndexes = list([v[2] for v in eventTable.values()])
        self.eventIndex_to_eventId = dict(zip(self.eventIndexes, self.eventIds))
        self.rowSize = (
            self._header.channels
            * self._config.samplesPerRow
            * self._header.bitDepth
            / 8
        )
        self.metadata = json.loads(str(metadata, "UTF-8"))
        self.event_well_info = self.get_well_info_by_evt()
        self.has_well_info = len(self.event_well_info) > 0
        self.get_phasors()

    def get_phasors(self):
        # used for phase correction
        def make_complex_array(complexlist):
            complexarray = np.empty(len(complexlist) // 2, dtype="complex64")
            for i in range(len(complexlist) // 2):
                complexarray[i] = complexlist[2 * i] + 1.0j * complexlist[2 * i + 1]
            return complexarray

        self.refPhasors = make_complex_array(self.metadata["lasers"][0]["refPhasors"])[
            16:120
        ]
        self.intrinsicPhasors = {}
        intPhasors = self.metadata["lasers"][0]["intrinsicPhasors"]
        for k in intPhasors.keys():
            try:
                self.intrinsicPhasors[k] = make_complex_array(intPhasors[k])
            except Exception as e:
                print(e)

    def get_well_info_by_evt(self):
        # for each event in an index sort file, get (wellId, populationName)
        event_well_info = {}
        if self.metadata["wells"] is None:
            return event_well_info
        for well in self.metadata["wells"]:
            for eventIndex in well["events"]:
                event_well_info[eventIndex] = (well["wellId"], well["populationName"])
        return event_well_info

    def get_eventId(self, relative_eventId):
        # get the absolute eventId
        # this is used to index into the eventTable to extract waveforms
        try:
            eventId = self.eventIds[relative_eventId]
        except IndexError as e:
            eventId = None
        return eventId

    def get_eventIndex(self, eventId):
        # get the relative eventIndex (relative to all FCS events, not relative to waveform events)
        # this matches the sequential event number in the FCS file
        try:
            eventIndex = self.eventTable[eventId][2]
        except KeyError as e:
            eventIndex = None
        return eventIndex

    def get_waveform(self, eventId, rel_id=False):
        # extract the raw waveform data
        try:
            if rel_id:
                eventId = self.get_eventId(eventId)
            offset, length, eventIndex = self.eventTable[eventId]
        except KeyError as e:
            raise EventIDNotFoundException("Event not found") from e
        with open(self.fpath, "rb") as f:
            f.seek(offset)
            waveform = f.read(length)
        return waveform

    def get_image(self, eventId, rel_id=False):
        # convert raw waveform data to an image stack using FFT
        # eventId - absolute event index, matches eventId in FCS
        # rel_id - use relative event index (starting at 0), for convenience

        # this should be checked against src/net/Projects/Chorus.Analysis.ImageGeneration/ImageGenerator.cs
        # and src/net/Projects/Chorus.Analysis/Algorithms/FireEngine.cs TryDecodingImageFromWaveform

        # grab waveform
        wave = self.get_waveform(eventId, rel_id)
        wave = np.frombuffer(wave[:-4096]).view("uint16").reshape(-1, self.num_channels)

        # compute the fast phase
        brightfield_channel = self.brightfield_channel
        first_row = wave[: self.num_per_row, brightfield_channel]
        first_fft = np.fft.rfft(first_row)[self.image_bin_min : self.image_bin_max]
        fast_phase = np.mean(first_fft / self.refPhasors)
        fast_phase /= np.abs(fast_phase)

        # compute raw ffts, and do sub-row phase correction
        # I'm sure there's a more efficient way to do this
        num_rows = (wave.shape[0] - self.num_per_row) // self.stride + 1
        stacked_ffts = np.empty(
            (num_rows, self.num_columns, self.num_channels), dtype="complex64"
        )
        offset = 0
        for i in range(num_rows):
            w = wave[offset : offset + self.num_per_row, :]
            stacked_ffts[i, :, :] = np.fft.rfft(w, axis=0)[
                self.image_bin_min : self.image_bin_max
            ]
            shift = offset % self.num_per_row
            lineshift = np.exp(
                shift * 2.0j * np.pi * np.linspace(0, 0.5, self.num_per_row // 2 + 1)
            )[self.image_bin_min : self.image_bin_max]
            stacked_ffts[i, :, :] /= lineshift[:, None]
            offset += self.stride

        # perform phase adjustment per channel
        stacked_image = None
        for channel in self.img_channels:
            im = stacked_ffts[:, :, channel]
            try:
                im /= self.intrinsicPhasors[str(channel)][None, :] * fast_phase
            except KeyError:
                pass
            im = np.real(im)

            if stacked_image is None:
                stacked_image = im.reshape((im.shape[0], im.shape[1], 1))
            else:
                stacked_image = np.concatenate(
                    (stacked_image, im[:, :, np.newaxis]), axis=2
                )

        stacked_image /= 65535 * 8.0

        return stacked_image

    def get_multipage_image_for_tiff(self, eventId, rel_id=False, **img_kwargs):
        # generate image and rearrange axes for tiff export
        im_stack = self.get_image(eventId, rel_id=rel_id, **img_kwargs)
        multipage_image = np.moveaxis(im_stack, 2, 0)
        return multipage_image

    def write_multipage_tiff(self, multipage_image, fpath_out):
        # write tiff to file
        with TiffWriter(fpath_out) as tif:
            for rel_ch_idx in range(len(multipage_image)):
                page = multipage_image[rel_ch_idx]
                extratags = self.get_tiff_tag(rel_ch_idx)
                tif.write(
                    page,
                    resolution=self.resolutions,
                    software=self.software,
                    description=None,
                    extratags=extratags,
                )

    def get_and_write_tiff(
        self, fdir_out, eventId, rel_id=False, isSort=False, img_kwargs=None, **name_kwargs
    ):
        # get tiff with given eventId and write to fdir_out
        img_kwargs = {} if img_kwargs is None else img_kwargs
        if rel_id:
            eventId = self.get_eventId(eventId)
        multipage_image = self.get_multipage_image_for_tiff(
            eventId, rel_id=False, **img_kwargs
        )
        fname = self.get_tiff_name(eventId, **name_kwargs)

        if not isSort:
            numPerFolder = 10000
            evt = self.get_eventIndex(eventId)
            evStr = str(evt)
            ev = math.floor(evt / numPerFolder) * numPerFolder
            sub = "%08d" % ev
            sub = sub.zfill(8)
            fdir_out = os.path.join(fdir_out, sub)
            if not os.path.exists(fdir_out):
                os.makedirs(fdir_out)

        fullFilePath = os.path.join(fdir_out, fname)
        if os.path.exists(fullFilePath):
            print('Found existing image file ' + fname)
            return multipage_image

        self.write_multipage_tiff(multipage_image, fullFilePath)
        return multipage_image

    def export_tiffs(
        self,
        fpath_out,
        eventIds,
        export_all=False,
        rel_id=False,
        is_eventIndex=True,
        return_images=False,
        verbose=True,
        **name_kwargs,
    ):
        # generate and save tiffs for the selected events
        # eventIds: array-like of events to fetch.
        #           if rel_id=True, these are relative sequential index of waveform recordings
        #           if rel_id=False and is_eventIndex=True, these are the eventIndex (relative sequential index of FCS events)
        #           if rel_id=False and is_eventIndex=False, these are absolute eventId (matched to FCS EventNumber0)
        # rel_id: bool flag indicates whether to use sequential relative image indexing
        # is_eventIndex: bool flag indicating to use "eventIndex" if rel_id is false. This is the eventIndex that FlowJo uses

        imgs = []
        maxEventIds = len(self.eventIds)
        offset = 0 #  maxEventIds / total_segments

        export_ct = 0
        failure_ct = 0
# REPLACE NEXT
        isSort = IS_SORT
        for ct, evt in enumerate(eventIds):
            if ct < maxEventIds:
                try:
                    if is_eventIndex and not rel_id:
                        # convert from eventIndex to eventId
                        try:
                            evt = self.eventIndex_to_eventId[evt]
                        except KeyError as e:
                            raise EventIDNotFoundException("Event not found") from e

                    img = self.get_and_write_tiff(
                        fpath_out, evt, rel_id=rel_id, isSort=isSort, **name_kwargs
                    )
                    export_ct += 1
                    if not ct % 1000:
                        print('Extracted ' + str(ct))

                except EventIDNotFoundException as e:
                    # if an event is not found, continue
                    failure_ct += 1
                    continue

                if return_images:
                    imgs.append(img)
            else:
                break

        if return_images:
            return {
                "images": imgs,
                "exportCount": export_ct,
                "failureCount": failure_ct,
            }
        else:
            return {
                "images": None,
                "exportCount": export_ct,
                "failureCount": failure_ct,
            }

    def get_tiff_name(
        self, eventId, use_recording_name=True, use_wellinfo=True, suffix=None
    ):
        eventIndex = self.get_eventIndex(eventId)
        names = []
        if use_recording_name:
            names += [self.metadata["recordingName"]]
        if use_wellinfo:
            if self.event_well_info.get(eventIndex, None) is not None:
                well, pop = self.event_well_info[eventIndex]
                names += [well, pop]
        names += [str(eventIndex).zfill(self.fname_zeros)]
        if suffix is not None:
            names += [suffix]

        name = "_".join(names)
        name = ".".join([name, self.tiff_ext])

        return name

    def get_horizontal_resolution(self):
        # see ImageGenerator.ScaleHorizontalResolution
        return self.resolution / (
            self.fireconstants_multiplicationfactor * self.pixel_aspect_ratio
        )

    def get_tiff_tag(self, rel_ch_idx):
        # refer to src/net/Projects/Chorus.Analysis.ImageGeneration/ImageGenerator.cs

        # output.SetField(TiffTag.IMAGEWIDTH, width); 256
        # output.SetField(TiffTag.IMAGELENGTH, height); 257
        # output.SetField(TiffTag.SAMPLESPERPIXEL, 1); 277
        # output.SetField(TiffTag.BITSPERSAMPLE, 32); 258
        # output.SetField(TiffTag.ORIENTATION, Orientation.TOPLEFT); 274
        # output.SetField(TiffTag.ROWSPERSTRIP, height); 278
        # output.SetField(TiffTag.SAMPLEFORMAT, SampleFormat.IEEEFP); 339 ***J
        # output.SetField(TiffTag.XRESOLUTION, 88.0); 282
        # output.SetField(TiffTag.YRESOLUTION, 88.0); 283
        # output.SetField(TiffTag.RESOLUTIONUNIT, ResUnit.CENTIMETER); 296 ***J
        # output.SetField(TiffTag.PLANARCONFIG, PlanarConfig.CONTIG); 284 ***J
        # output.SetField(TiffTag.PHOTOMETRIC, Photometric.MINISBLACK); 262
        # output.SetField(TiffTag.COMPRESSION, Compression.NONE); 259
        # output.SetField(TiffTag.FILLORDER, FillOrder.MSB2LSB); 266 ***J
        #
        # void writeMetadata(Tiff tiff, int frameIndex)
        #  {
        #     tiff.SetField(TiffTag.PAGENAME, mapping[frameIndex]); #285

        #     // For tiff images X and Y RESOLUTION tags are used to define horizontal and vertical pixel density (DPI - dot per inch).
        #     // Their ratio represents pixel aspect ratio of image, so for non square pixels (which we have) ratio is different than 1.
        #     // Images need to be stretch horizontally by the specified ratio & factor (or the height needs to be divided by ratio & factor).
        #     tiff.SetField(TiffTag.XRESOLUTION, ScaleHorizontalResolution(Resolution, pixelAspectRatio)); #282
        #     tiff.SetField(TiffTag.YRESOLUTION, Resolution); #283
        #  }
        #
        # ScaleHorizontalResolution... resolution / (pixelAspectRatio.Value * FireConstants.MultiplicationFactor);
        # FireConstants.MultiplicationFactor = 4

        # For tifffile, we format tags as a 5-tuple with (tag code, datatype code, number of values, value, apply only to first page)
        # Tags 256, 257, 258, 259, 262, 270, 273, 277, 278, 279 are set by tifffile automatically based on image
        tags = [
            # (266, 3, 1, 1, False),  # FillOrder ***J
            (274, 3, 1, 1, False),  # Orientation
            # (282, 5, 1, self.get_horizontal_resolution(), False), #XResolution
            # (283, 5, 1, self.resolution, False), #YResolution
            # (284, 3, 1, 1, False),  # PlanarConfiguration ***J
            (285, 2, 1, self.img_channel_names[rel_ch_idx], False),  # PageName
            # (296, 3, 1, 3, False),  # ResolutionUnit ***J
            # (339, 3, 1, 3),  # SampleFormat ***J
        ]
        return tags

    class Header:
        def __init__(self):
            pass

class EventIDNotFoundException(Exception):
    pass

try:
# REPLACE NEXT
    cvwPath = r"TEST_EVENTS_FILE"
    waveformReader = WaveformFileReader(cvwPath)
# REPLACE NEXT
    imagesFolder = 'IMAGE_DIR'
# REPLACE NEXT
    numEvts = EXPLAIN_EVENT_NUM
    waveformReader.export_tiffs(imagesFolder,range(numEvts),rel_id=True)
    
except Exception as inst:
  print('Exception in extraction script')
  print(type(inst))
  print(inst.args)
  print(inst)

else:
  print('Extraction script completed')

finally:
# REPLACE NEXT
  outFile = open("PROBABILITIES_OUT_FILE" + ".done", "w")
  outFile.close()
    